@@ -2,6 +2,7 @@ from typing import Any, Optional, Type, Union
 
 import gym
 import numpy as np
+from gym import spaces
 from gym.envs.registration import register
 from mathy_core.rule import ExpressionChangeRule
 
@@ -9,7 +10,6 @@ from ..env import MathyEnv
 from ..state import MathyEnvState, MathyObservation
 from ..time_step import is_terminal_transition
 from ..types import ActionType, MathyEnvProblemArgs
-from ..util import pad_array
 from .masked_discrete import MaskedDiscrete
 
 
@@ -22,6 +22,7 @@ class MathyGymEnv(gym.Env):
     _challenge: Optional[MathyEnvState]
     env_class: Type[MathyEnv]
     env_problem_args: Optional[MathyEnvProblemArgs]
+    mask_as_probabilities: bool
 
     def __init__(
         self,
@@ -29,11 +30,13 @@ class MathyGymEnv(gym.Env):
         env_problem_args: Optional[MathyEnvProblemArgs] = None,
         env_problem: Optional[str] = None,
         env_max_moves: int = 64,
-        np_observation: bool = False,
+        np_observation: bool = True,
+        mask_as_probabilities: bool = False,
         repeat_problem: bool = False,
         **env_kwargs: Any,
     ):
         self.state = None
+        self.mask_as_probabilities = mask_as_probabilities
         self.repeat_problem = repeat_problem
         self.np_observation = np_observation
         self.mathy = env_class(**env_kwargs)
@@ -45,15 +48,38 @@ class MathyGymEnv(gym.Env):
             )
         else:
             self._challenge, _ = self.mathy.get_initial_state(env_problem_args)
+
         self.action_space = MaskedDiscrete(self.action_size, [1] * self.action_size)
+        if np_observation is True:
+            mask = len(self.mathy.rules) * self.mathy.max_seq_len
+            values = self.mathy.max_seq_len
+            nodes = self.mathy.max_seq_len
+            type = 2
+            time = 1
+            obs_size = mask + values + nodes + type + time
+            self.observation_space = spaces.Box(
+                low=0, high=1, shape=(obs_size,), dtype="float32"
+            )
+        else:
+            # TODO: How to express a structured observation space in gym? e.g.
+            #
+            # obs = (
+            #      nodes_tensor(1, 128),
+            #      values_tensor(1, 128),
+            #      type_tensor(1, 2),
+            #      time_tensor(1, 1)
+            # )
+            raise NotImplementedError(
+                "not sure how to represent the mathy obs space in gym"
+            )
 
     @property
     def action_size(self) -> int:
-        if self.state is not None:
-            return self.mathy.get_agent_actions_count(self.state)
         return self.mathy.action_size
 
-    def step(self, action: ActionType) -> Union[MathyObservation, np.ndarray]:
+    def step(
+        self, action: Union[int, ActionType]
+    ) -> Union[MathyObservation, np.ndarray]:
         assert self.state is not None, "call reset() before stepping the environment"
         self.state, transition, change = self.mathy.get_next_state(self.state, action)
         done = is_terminal_transition(transition)
@@ -70,17 +96,22 @@ class MathyGymEnv(gym.Env):
         """Observe the environment at the given state, updating the observation
         space and action space for the given state."""
         action_mask = self.mathy.get_valid_moves(state)
-        observation = self.mathy.state_to_observation(state)
-        self.action_space.n = self.mathy.get_agent_actions_count(state)
-        self.action_space.mask = action_mask
-        if self.np_observation:
-            # convert mask to probabilities
-            nodes = np.array(pad_array(observation.nodes, 512, 0))
-            mask = np.array(pad_array(observation.mask, 512, 0))
-            mask_sum = np.sum(mask)
+        observation = self.mathy.state_to_observation(
+            state, max_seq_len=self.mathy.max_seq_len
+        )
+        flat_mask = np.array(action_mask).reshape(-1)
+        self.action_space.update_mask(flat_mask)
+        if self.mask_as_probabilities:
+            mask_sum = np.sum(flat_mask)
             if mask_sum > 0.0:
-                mask = mask / mask_sum
-            return np.vstack((nodes, mask))
+                flat_mask = flat_mask / mask_sum
+        if self.np_observation:
+            nodes = np.array(observation.nodes, dtype="float32").reshape(-1)
+            values = np.array(observation.values, dtype="float32").reshape(-1)
+            np_observation = np.concatenate(
+                [observation.type, observation.time, nodes, values, flat_mask], axis=-1
+            )
+            return np_observation
         return observation
 
     def reset(self) -> Union[MathyObservation, np.ndarray]:
